#
# Revision: 3.0.180532
#
# Copyright (c) 2018 Dell Technologies Inc.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# DELL EMC CORPORATION. USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF DELL EMC.
#
package SCAgentState;
use lib qw(. ./sc);
use strict;
use warnings;
use parent 'ISerializable';
use SCTools qw(ISIN);
use SCLogger;
use SCAgentVPair;


use constant {

    #GENERAL
    STATUS_UNKNOWN              => 0,

    #CONFIGURATION states
    STATUS_CONFIG_AGENT_INVALID => 100,
    STATUS_CONFIG_DISCONNECTED  => 110,
    STATUS_CONFIG_MISONFIGURED  => 120,
    STATUS_CONFIG_HEALTHY_EMPTY => 130,
    STATUS_CONFIG_HEALTHY       => 140, #== LEVEL_CONFIGURED

    #REPLICATION states
    STATUS_RPOINT_UNHEALTHY             => 209,
    STATUS_RPOINT_RUNNINGGOOD_MIXED     => 220, #== LEVEL_STABLE, LEVEL_RUNNINGGOOD
    STATUS_RPOINT_RUNNINGGOOD_REVERSE   => 230, #== LEVEL_STABLE, LEVEL_RUNNINGGOOD
    STATUS_RPOINT_RUNNINGGOOD_DIRECT    => 240, #== LEVEL_STABLE, LEVEL_RUNNINGGOOD

    #WRITE ACCESS states
    STATUS_WRITEACCESS_CONSISTENT       => SCAgentVPair::MAPPING_CONSISTENT,
    STATUS_WRITEACCESS_ACCESSIBLE       => SCAgentVPair::MAPPING_WRITEACCESS,
    STATUS_WRITEACCESS_EXCLUSIVE        => SCAgentVPair::MAPPING_EXCLUSIVE,

    #READONLY ACCESS states
    STATUS_READONLY_NOT_CONFIRMED       => 400,
    STATUS_READONLY_CONFIRMED           => 401
};



use constant {

    LEVEL_RUNNINGGOOD       => STATUS_RPOINT_RUNNINGGOOD_MIXED()
};



use constant {

    PROPERTY_AGENT          => 'AGENT_STATE_SCAGENT',
    PROPERTY_CONFIGURATION  => 'AGENT_STATE_CONFIGURATION',
    PROPERTY_RESTOREPOINTS  => 'AGENT_STATE_RESTOREPOINTS',
    PROPERTY_WRITEACCESS    => 'AGENT_STATE_WRITEACCESS',
    PROPERTY_READONLY       => 'AGENT_STATE_READONLY',
    PROPERTY_ISSUES         => 'AGENT_STATE_ISSUES'
};



sub new {

    my              $class  = shift;
    my              $agent  = shift;

    return bless {

            PROPERTY_AGENT()            => $agent,
            PROPERTY_CONFIGURATION()    => 0,
            PROPERTY_RESTOREPOINTS()    => 0,
            PROPERTY_WRITEACCESS()      => 0,
            PROPERTY_ISSUES()           => []

    }, $class;
}





#@method
#@override
sub properties      { qw (name ssn) }

#@method
#@returns SCAgent
sub agent           { $_[0]->{PROPERTY_AGENT()} }

#@method
#@returns sc::api
sub api             { $_[0]->agent->api }


#@method
sub configState     { $_[0]->{PROPERTY_CONFIGURATION()} = $_[0]->_check_configuration   unless $_[0]->{PROPERTY_CONFIGURATION()};   return $_[0]->{PROPERTY_CONFIGURATION()} }


#@method
sub rpointState     { $_[0]->{PROPERTY_RESTOREPOINTS()} = $_[0]->_check_replications    unless $_[0]->{PROPERTY_RESTOREPOINTS()};   return $_[0]->{PROPERTY_RESTOREPOINTS()} }


#@method
sub accessState     { $_[0]->{PROPERTY_WRITEACCESS()}   = $_[0]->_check_writeaccess     unless $_[0]->{PROPERTY_WRITEACCESS()};     return $_[0]->{PROPERTY_WRITEACCESS()} }


#@method
sub confirmReadOnly { $_[0]->{PROPERTY_READONLY()}      = $_[0]->_confirm_readonly      unless $_[0]->{PROPERTY_READONLY()};        return $_[0]->{PROPERTY_READONLY()} }


#@method
sub isEmpty         { $_[0]->configState == STATUS_CONFIG_HEALTHY_EMPTY() }


#@method
sub isConfigured    { $_[0]->configState == STATUS_CONFIG_HEALTHY() }


#@method
sub isRpointHealthy { $_[0]->rpointState >= STATUS_RPOINT_RUNNINGGOOD_MIXED() }


#@method
sub isMapConsistent { $_[0]->accessState & STATUS_WRITEACCESS_CONSISTENT() }


#@method
sub isMapExclusive  { $_[0]->accessState & STATUS_WRITEACCESS_EXCLUSIVE() }


#@method
sub isWriteEnabled  { $_[0]->accessState & STATUS_WRITEACCESS_ACCESSIBLE() }


#@method
sub isReadOnly      { $_[0]->confirmReadOnly == STATUS_READONLY_CONFIRMED() }







#@method
sub isSafeToOnline  {

    my $self = shift;
    my $result = $self->isRpointHealthy && $self->isMapExclusive;

    TRACE '[SCAGENT][SAFE-TO-GO-ONLINE]', {

            'rpoints are healthy'       => $self->isRpointHealthy,
            'mapping is exclusive'      => $self->isMapExclusive,
            'mapping is consistent'     => $self->isMapConsistent,
            'mapping is write-enabled'  => $self->isWriteEnabled,
            'verdict'                   => $result
        },  'rpoints are healthy', 'mapping is exclusive', 'mapping is consistent', 'mapping is write-enabled', 'verdict';

    return $result;
}







#@method
sub isOnline {

    my $self = shift;
    if($self->agent->areParentsOnline) {

        TRACE '[SCAGENT-STATE][IS ONLINE]', 'All Parents are ONLINE => should keep the agent online';
        return 1
    }

    my $result = 0;

    if($self->isRpointHealthy && $self->isWriteEnabled) {

        $result = 1;
    }

    TRACE '[SCAGENT-STATE][IS ONLINE]', {
            'parents'                   => 'some parents are not online',
            'rpoints are healthy'       => $self->isRpointHealthy,
            'mapping is exclusive'      => $self->isMapExclusive,
            'mapping is consistent'     => $self->isMapConsistent,
            'mapping is write-enabled'  => $self->isWriteEnabled,
            'verdict'                   => $result
        };

    return $result;
}







#@method
sub isOffline {

    my $self = shift;
    my $result = not $self->isWriteEnabled;

    TRACE '[SCAGENT-STATE][IS ONLINE]', {
            'parents'                   => 'some parents are not online',
            'rpoints are healthy'       => $self->isRpointHealthy,
            'mapping is exclusive'      => $self->isMapExclusive,
            'mapping is consistent'     => $self->isMapConsistent,
            'mapping is write-enabled'  => $self->isWriteEnabled,
            'verdict'                   => $result
        };

    return not $self->isWriteEnabled;
}







#@method
sub _check_configuration {

    my $self = shift;
    my $agent = $self->agent;
    return STATUS_CONFIG_AGENT_INVALID() unless $agent->isValid;

    my @volumes = $agent->volumes;
    my @servers = $agent->servers;
    my @nodes   = $agent->localNodes;

    return STATUS_CONFIG_HEALTHY_EMPTY() unless @volumes || @servers || @nodes;
    return STATUS_CONFIG_MISONFIGURED()  unless @volumes && @servers && @nodes;
    return STATUS_CONFIG_DISCONNECTED()  unless $agent->connect;

    if (my @unhealthy = grep { not $_->isHealthy } @servers) {

        LOG MSG_AGENT_UNHEALTHY_SERVERS, @unhealthy;
        return STATUS_CONFIG_MISONFIGURED();
    }

    return STATUS_CONFIG_HEALTHY();
}







#@method
sub _check_replications {

    my $self    = shift;
    my $agent   = $self->agent;
    my @volumes = $agent->volumes;

    if ($self->isConfigured) {

        if (my @unhealthy = grep {not $_->isHealthy} @volumes) {

            LOG MSG_AGENT_UNHEALTHY_REPLICA, @unhealthy;
            return STATUS_RPOINT_UNHEALTHY();
        }

        my $nCount   = scalar(@volumes);
        my $nDirect  = 0;
        my $nReverse = 0;

        foreach my SCAgentVPair $vp (@volumes) {

            my sc::restorepoint $rp = $vp->rpoint;
            $nDirect++  if !$rp->isDRActivated && $vp->isLocal($rp);
            $nDirect++  if  $rp->isDRActivated && $vp->isRemote($rp);
            $nReverse++ if !$rp->isDRActivated && $vp->isRemote($rp);
            $nReverse++ if  $rp->isDRActivated && $vp->isLocal($rp);
        }

        return STATUS_RPOINT_RUNNINGGOOD_DIRECT()  if $nCount == $nDirect;
        return STATUS_RPOINT_RUNNINGGOOD_REVERSE() if $nCount == $nReverse;

        LOG MSG_AGENT_MIXED_DIRECTIONS, { agent => $agent->name }, @volumes;
        return STATUS_RPOINT_RUNNINGGOOD_MIXED();
    }

    return STATUS_UNKNOWN();
}







#@method
sub _check_writeaccess {

    my $self    = shift;
    my $agent   = $self->agent;
    my @volumes = $agent->volumes;
    my $result  = STATUS_UNKNOWN();

    if ($self->isConfigured) {

        $result = STATUS_WRITEACCESS_ACCESSIBLE() | STATUS_WRITEACCESS_CONSISTENT() | STATUS_WRITEACCESS_EXCLUSIVE();

        foreach my SCAgentVPair $vp (@volumes) {

            $result &= $vp->classifyMappings();
            last unless $result;
        }
    }

    TRACE '[SCAGENT-STATE][MAPPINGS CHECK][VPAIRS]', [@volumes];
    TRACE '[SCAGENT-STATE][MAPPINGS CHECK][RESULT]',  $result;
    return $result;
}







#@method
sub _confirm_readonly {

    my $self    = shift;
    my $agent   = $self->agent;
    my @volumes = $agent->volumes;
    my $result  = 0;

    if ($self->isConfigured) {

        foreach my SCAgentVPair $vp (@volumes) {

            my $av = $vp->activeVolume;
            my @maps = grep { $self->agent->isCurrentHost($_->serverID) } $av->mappings;
            my @ro   = grep { $_->isRO } @maps;
            my @rw   = grep {!$_->isRO } @maps;
               $result = 1;
               $result = 0 if @rw;
               $result = 0 unless @ro;

            TRACE '[SCAGENT-STATE][READONLY DOUBLE CHECK][CONFIRMATION]', [$vp, @maps] if $result;
            TRACE '[SCAGENT-STATE][READONLY DOUBLE CHECK][FAILURE TO CONFIRM READONLY]', [$vp, @maps] unless $result;
            last if $result;
        }
    }

    TRACE '[SCAGENT-STATE][READONLY DOUBLE CHECK][RESULT]', ($result ? 'Confirmed ReadOnly' : 'Could not confirm');
    return $result ? STATUS_READONLY_CONFIRMED() : STATUS_READONLY_NOT_CONFIRMED();
}



1;