#
# Revision: 3.0.180532
#
# Copyright (c) 2018 Dell Technologies Inc.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# DELL EMC CORPORATION. USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF DELL EMC.
#
package SCAgentVPair;
use strict;
use warnings;
use parent 'ISerializable';
use SCTools qw(SAFEVALUE SAFEKEY DATETIME ISIN GREP UNIQ GREPUNIQ TOSTRING);
use SCAgent;
use SCLogger;





use constant {

    VPAIR_Key       => 'key',
    VPAIR_IDs       => 'id',
    VPAIR_Serials   => 'serials',
    VPAIR_Rpoint    => 'rpoint',
    VPAIR_Agent     => 'agent',
    VPAIR_Task      => 'task',
    VPAIR_Swap      => 'swap',
    VPAIR_Retry     => 'retry',
    VPAIR_Log       => 'log',
    VPAIR_Status    => 'status'
};





use constant {

    HEALTH_OK_LV_UP             => 'volume is protected (livevolume is UP)',
    HEALTH_OK_RP_UP             => 'volume is protected (replication is UP)',
    HEALTH_OK_AT_DR_SITE        => 'source site is DOWN, but we are on Remote Site',
    HEALTH_OK_DRACTIVATED       => 'volume pair is OK (livevolume is DrActivated)',
    HEALTH_OK_RPOINT_DEFINED    => 'volume pair is OK (rpoint is in defined healthy state)',
    HEALTH_ISSUE_SWAP           => 'volume pair needs to reverse its replication direction (swap is registered but is not completed)',
    HEALTH_ISSUE_SOURCE_FAILED  => 'source site is DOWN and we need to go away from here',
    HEALTH_ISSUE_AV_DOWN        => 'volume is UNPROTECTED (active volume is DOWN)',
    HEALTH_ISSUE_AV_INVALID     => 'volume is UNPROTECTED (failure to identify active volume)',
    HEALTH_ISSUE_LV_AFO         => 'volume is UNPROTECTED (livevolume AutoFailover feature is enabled which is not supported with host-based cluster agent); please, switch off AutoFailover feature or use LiveVolume AutoFailover as a self-sufficient clustering out of control of this agent',
    HEALTH_ISSUE_LV_STATE       => 'volume is UNPROTECTED (livevolume is not running)',
    HEALTH_ISSUE_LV_HCMODE      => 'volume is UNPROTECTED (livevolume uses sync replication in the HighConsistency mode which prevents failover capabilities); please change livevolume replication mode to HighAvailability)',
    HEALTH_ISSUE_LV_INVALID     => 'volume is UNPROTECTED (livevolume is invalid)',
    HEALTH_ISSUE_REPL_DOWN      => 'replication is not running',
    HEALTH_ISSUE_REPL_HCMODE    => 'replication is in HighConsistency mode which prevents failover capabilities; please change sync replication mode to HighAvailability',
    HEALTH_ISSUE_REPL_STATE     => 'volume is UNPROTECTED (bad replication state)',
    HEALTH_ISSUE_REPL_MISSING   => 'volume is UNPROTECTED (replication is missing)',
    HEALTH_ISSUE_RPOINT_UNDEF   => 'volume is UNPROTECTED (restore point is in undefined state)'
};





use constant {

    MAPPING_CONSISTENT  => 1,
    MAPPING_WRITEACCESS => 2,
    MAPPING_EXCLUSIVE   => 4
};





sub new {

    my                  $class  = shift;
    my SCAgent          $agent  = shift;
    my sc::restorepoint $rpoint = shift;
    my sc::volume       $R1     = $rpoint->sourceVolume;
    my sc::volume       $R2     = $rpoint->targetVolume;
    my sc::volume       $DR     = $rpoint->isDRActivated ? $rpoint->activatedVolume : 0;
    my                  @vset   = grep { $_->isValid } grep { $_ } ($R1, $R2, $DR);

    return bless {
                        VPAIR_Agent()   => $agent,
                        VPAIR_Serials() => [ UNIQ(map { $_->serial } @vset) ],
                        VPAIR_IDs()     => [ UNIQ(map { $_->id     } @vset) ],
                        VPAIR_Key()     => SAFEKEY($R1->name).' ('.$R1->id.')',
                        VPAIR_Rpoint()  => $rpoint->id,
                        VPAIR_Task()    => 0,
                        VPAIR_Swap()    => 0,
                        VPAIR_Retry()   => 0,
                        VPAIR_Log()     => []
        }, $class;
}





sub _decode {

    my ($class, $agent, $key, $body) = @_;
    my  $proto = JSON::from_json ($body);
        $proto->{VPAIR_Agent()} = $agent;
        $proto->{VPAIR_Key()}   = $key;
        $proto->{VPAIR_Task()}  = 0;
        $proto->{VPAIR_Retry()} = 0;
        $proto->{VPAIR_Log()}   = [];

    return bless $proto, $class;
}





sub _encode {

    my  $self = shift;
    my  $proto = { map { $_ => ($self->{$_} || 0) } (VPAIR_IDs,
                                                     VPAIR_Serials,
                                                     VPAIR_Rpoint,
                            ($self->{VPAIR_Swap()} ? VPAIR_Swap : ())) };

    return JSON::to_json($proto);
}





#@method
sub update {

    my  $self = shift;
    my  sc::restorepoint $rp = shift;

    if ($rp && $rp->isValid) {

        $rp = $rp->validate if $rp->isNeedValidation;

        if ($rp->isDefinedState) {

            my sc::volume   $R1 = $rp->sourceVolume;
            my sc::volume   $R2 = $rp->targetVolume;
            my sc::volume   $DR = $rp->isDRActivated ? $rp->activatedVolume : 0;
            my @vset = grep {$_->isValid} grep {$_} ($R1, $R2, $DR);

            $self->{VPAIR_Rpoint()} = $rp->id;
            $self->{VPAIR_Serials()} = [ UNIQ(map {$_->serial} @vset) ];
            $self->{VPAIR_IDs()} = [ UNIQ(map {$_->id} @vset) ];
            $self->{VPAIR_Swap()} = 0;
            $self->log('update', 'writing new config', (my $r = $self->agent->updateVpair($self)) ? 'OK' : 'FAILED');
            return $r;
        }

        else {

            my $old = $self->rpid;
            $self->{VPAIR_Rpoint()} = $rp->id;
            $self->log('update', 'rpoint is not in a defined state; only rpid will be updated temporarily: ' . $old . ' -> ' . $self->rpid);
        }
    }

    return 0;
}

#@method
#@override
sub properties  { qw (key serials vids rpid status) }

#@method
#@returns sc::api
sub api         { $_[0]->agent->api }

#@method
#@returns SCAgent
sub agent       { $_[0]->{VPAIR_Agent()} }


#@method
sub serials     {(@{$_[0]->{VPAIR_Serials()}||[]}, ($_[0]->{VPAIR_Swap()}||())) }


#@method
sub vids        { @{$_[0]->{VPAIR_IDs()}||[]} }


#@method
sub rpid        { $_[0]->{VPAIR_Rpoint()}||'' }


#@method
sub key         { $_[0]->{VPAIR_Key()}  || ($_[0]->{VPAIR_Key()} = SAFEKEY($_[0]->rpoint->sourceVolume->name)) }


#@method
sub swap        { $_[0]->{VPAIR_Swap()} || 0 }


#@method
sub isLocal     { my $self = shift; my sc::restorepoint $rp = shift || $self->rpoint; return ISIN($rp->sourceSSN, $self->agent->localNodes) }


#@method
sub isRemote    { my $self = shift; my sc::restorepoint $rp = shift || $self->rpoint; return ISIN($rp->targetSSN, $self->agent->localNodes) }


#@method
sub history     { my $self = shift; return join("\n", @{$self->{VPAIR_Log()}||[]}) }


#@method
sub trace       { my ($self, $func, $msg, @info) = @_; TRACE '['.$self->key.'][ '.$func.' ] ' . $msg, (1 < @info ? [@info] : @info) }


#@method
sub status      { my ($self, $issue) = @_; $self->{VPAIR_Status()} = join("; ", ($issue, $self->{VPAIR_Status()}||())) if $issue; return $self->{VPAIR_Status()} || 'not checked' }


#@method
sub log         {

    my      $self = shift;
    my      $func = shift || '#';
    my      $msg  = shift || '';
    push  @{$self->{VPAIR_Log()}}, (DATETIME().' - '.$func.': '.(@_ ? $msg.', '.TOSTRING($_[0]) : $msg));
            $self->trace($func, $msg, $self, @_);
}


#@method
sub logAndRetry {

    my  $self = shift;
        $self->log(@_);

    return 0 if $self->retry;

        $self->log(($_[0] || '#'), 'give up (retry limit: '.$self->{VPAIR_Retry()}.')');

    return 1;
}


#@method
sub isRescanRecommended {

    my $self    = shift;
    my $vxdisks = shift || { SCAgent::VXDISKS_MAP() };

    if (grep {exists($vxdisks->{$_})} $self->serials) {

        return 0 if $self->rpoint->isLiveVolume;
    }

    return 1;
}







#@method
sub classifyMappings {

    my $self = shift;
    my sc::volume       $av = shift || 0;
    my sc::restorepoint $rp = shift || $self->rpoint;

    unless ($rp && $rp->isValid) {

        $self->log('classify_mapping', 'rpoint is not valid; result = 0', $rp);
        return 0;
    }

    unless ($av) {

        if ($rp->isDefinedState) {

            $av = $self->activeVolume($rp);
        }
        else {

            $self->log('classify_mapping', 'rpoint is not in one of the Defined States; will try to use SourceVolume if it is visible on the host', $rp);
            my $vxdisks = { SCAgent::VXDISKS_MAP() };

            unless (grep {exists($vxdisks->{$_})} $self->serials) {

                $self->log('classify_mapping', 'volume is not visible on the host');
                return 0;
            }

            if ($self->isLocal($rp)) {

                $av = $rp->sourceVolume;
                $self->log('classify_mapping', 'volume is visible, and is Primary; will use the SourceVolume as active volume', $av);
            }
            else {

                $self->log('classify_mapping', 'volume is visible, but is not Primary', [ $self->agent->localNodes ], $rp);
                return 0;
            }
        }
    }

    if ($av) {

        my @rwmaps  = $self->activeMappingsRW($rp, $av);
        my $enabled = scalar( grep { $self->agent->isCurrentHost($_->serverID) } grep { $av->id eq $_->volumeID } @rwmaps );

        my $result = 0;
           $result |= MAPPING_WRITEACCESS() if $enabled;
           $result |= MAPPING_CONSISTENT()  if scalar(@rwmaps) <= 1;
           $result |= MAPPING_EXCLUSIVE()   if scalar(@rwmaps) == 0;
           $result |= MAPPING_EXCLUSIVE()   if scalar(@rwmaps) == 1 && $enabled;

        $self->log('classify_mapping', 'active volume: ' . ($enabled ? 'WRITE-ENABLED on this host' : 'NOT write-enabled on this host'));
        $self->log('classify_mapping', 'verdict: ' . $result);

        return $result;
    }

    $self->log('classify_mapping', 'failure to identify active volume; result = 0', $rp);
    return 0;
}







#@method
sub isHealthy {

    my $self = shift;
    my sc::restorepoint $rpoint = $self->rpoint;

    if ($self->swap) {

        $self->status(HEALTH_ISSUE_SWAP);
        $self->log('healthcheck', HEALTH_ISSUE_SWAP.': '.$self->swap.')');
        return 0;
    }


    if ($rpoint->isErrorState) {

        $self->log('healthcheck', 'rpoint is in error state ('.$rpoint->state.'), trying to reset rpoint', $rpoint);
        $self->resetRPoint($rpoint);
        $rpoint = $self->findRPoint;
    }


    if ($rpoint->isSourceFailed) {

        if ($self->isRemote($rpoint)) {

            $self->status(HEALTH_OK_AT_DR_SITE);
            $self->log('healthcheck', HEALTH_OK_AT_DR_SITE);
            return 1;
        }

        $self->status(HEALTH_ISSUE_SOURCE_FAILED);
        $self->log('healthcheck', HEALTH_ISSUE_SOURCE_FAILED);
    }

    elsif ($rpoint->isDefinedState) {

        my sc::volume $volume = $self->activeVolume($rpoint);

        if ($volume->isUp) {

            if ($rpoint->isRunningGood) {

                if ($rpoint->isLiveVolume) {

                    my sc::livevolume $lv = $rpoint->liveVolume;

                    if ($lv->isValid) {

                        if ($lv->isSyncHC) {

                            $self->status(HEALTH_ISSUE_LV_HCMODE);
                            $self->log('healthcheck', HEALTH_ISSUE_LV_HCMODE, $lv);
                        }
                        else {

                            if ($lv->isDrActivated) {

                                $self->status(HEALTH_OK_DRACTIVATED);
                                $self->log('healthcheck', HEALTH_OK_DRACTIVATED);
                                return 1;
                            }

                            if ($lv->isUp) {

                                $self->status(HEALTH_OK_LV_UP);
                                $self->log('healthcheck', HEALTH_OK_LV_UP);
                                return 1;
                            }

                            $self->status(HEALTH_ISSUE_LV_STATE.'; LiveVolume status: ['.(join('; ', $lv->status, $lv->replicaState, $lv->primaryStatus, $lv->secondaryStatus)).']');
                            $self->log('healthcheck', HEALTH_ISSUE_LV_STATE, $lv, $rpoint, $volume);
                        }
                    }

                    else {

                        $self->status(HEALTH_ISSUE_LV_INVALID);
                        $self->log('healthcheck', HEALTH_ISSUE_LV_INVALID, $lv, {%$volume}, $rpoint, $volume);
                    }
                }

                else {

                    my  @replicas = $volume->replications;

                    if (@replicas > 0) {

                        my $rcode = 1;

                        foreach my sc::replication $repl (@replicas) {

                            unless ($repl->isUp) {

                                $self->status(HEALTH_ISSUE_REPL_DOWN.' ['.$repl->state.'; '.$repl->status.']');
                                $self->log('healthcheck', HEALTH_ISSUE_REPL_DOWN, $repl, $rpoint, $volume);
                                $rcode = 0;
                            }

                            if ($repl->isSyncHC) {

                                $self->status(HEALTH_ISSUE_REPL_HCMODE);
                                $self->log('healthcheck', HEALTH_ISSUE_REPL_HCMODE, $repl);
                                $rcode = 0;
                            }
                        }

                        if ($rcode) {

                            $self->status(HEALTH_OK_RP_UP);
                            $self->log('healthcheck', HEALTH_OK_RP_UP);
                            return 1;
                        }

                        else {

                            $self->status(HEALTH_ISSUE_REPL_STATE);
                            $self->log('healthcheck', HEALTH_ISSUE_REPL_STATE, @replicas, $rpoint, $volume);
                        }
                    }

                    else {

                        $self->status(HEALTH_ISSUE_REPL_MISSING);
                        $self->log('healthcheck', HEALTH_ISSUE_REPL_MISSING, $rpoint, $volume);
                    }
                }
            }

            else {

                $self->status(HEALTH_OK_RPOINT_DEFINED.': '.$rpoint->state);
                $self->log('healthcheck', HEALTH_OK_RPOINT_DEFINED.': '.$rpoint->state, $rpoint);
                return 1;
            }
        }

        else {

            if ($volume->isValid) {

                $self->status(HEALTH_ISSUE_AV_DOWN);
                $self->log('healthcheck', HEALTH_ISSUE_AV_DOWN, $rpoint, $volume);
            }

            else {

                $self->status(HEALTH_ISSUE_AV_INVALID);
                $self->log('healthcheck', HEALTH_ISSUE_AV_INVALID, $rpoint, $volume) unless $volume->isValid
            }
        }
    }

    else {

        $self->status(HEALTH_ISSUE_RPOINT_UNDEF);
        $self->log('healthcheck', HEALTH_ISSUE_RPOINT_UNDEF, $rpoint);
    }

    return 0;
}







#@method
sub retry {

    my      $self   = shift;
    my      $limit  = shift || 5;
    my      $try    = ++$self->{VPAIR_Retry()};
    LOG     $try    < $limit ? MSG_VPAIR_RETRY : MSG_VPAIR_RETRY_LIMIT, {attempt => $try, limit => $limit};
    return  $try    < $limit;
}







#@method
sub prepareNewJob {

    my  $self    = shift;
        $self->{VPAIR_Retry()} = 0;
        $self->{VPAIR_Log()} = [];
        $self->rpoint->validate
}







#@method
#@returns sc::volume
sub volumes {

    my  $self = shift;
    my  @vset = $self->api->volumes($self->serials);
    return @vset if @vset;

    $self->log('get_volumes', 'failure to find volumes using my serial numbers', $self->serials);
    return ();
}







#@method
#@returns sc::restorepoint
sub rpoint {

    my $self = shift;

    if ($self->{VPAIR_Rpoint()}) {

        my      $rp = $self->api->get_rpoint($self->{VPAIR_Rpoint()});
                $rp = $rp->validate if $rp->isNeedValidation;
        return  $rp if $rp->isValid;
    }

    return $self->findRPoint;
}







#@method
#@returns sc::restorepoint
sub findRPoint {

    my  $self = shift;
        $self->log('find rpoint', 'need to re-determine my restore point', $self->{VPAIR_Rpoint()});

    my  @vset = ($self->volumes);
    my  @ids  = (grep {$_} (@vset ? (map { $_->id } @vset) : ($self->vids))); #Volume IDs are used only as a backup option; Serials are the primary way to re-determine RPoint reference
    my  @ssns = (map  {$_->ssn} @vset);
    my  $isLV = scalar(grep { $_->isLiveVolume } @vset);

    if (@ids) {

        unless ($isLV) {

            foreach my sc::replication $repl (map { $_->replications } @vset) {

                my sc::restorepoint $rp = $repl->restorepoint;

                unless ($rp->isValid) {

                    $self->log('find rpoint', 'replication is found, but restore point is not ready; saving new restore point', $repl->saveRPoint);
                    $rp = $repl->restorepoint;
                }

                if ($rp->isValid) {

                    $rp = $rp->validate if $rp->isNeedValidation;
                    $self->log('find rpoint', 'rpoint is found by volume replica', $rp);
                    $self->update($rp);
                    return $rp;
                }

                $self->log('find rpoint', 'failure to find valid rpoint using volume replica', $rp);
            }
        }


        my  @rpoints = GREP([qw(sourceVolumeID targetVolumeID activatedVolumeID)], [@ids], (grep { $_->isValid } $self->api->restorePoints(@ssns)));


        if ($isLV && not @rpoints) {

            foreach my sc::livevolume $lv (GREP([qw(primaryID secondaryID)], [@ids], (grep { $_->isValid } $self->api->livevolumes(@ssns)))) {

                $self->log('find rpoint', 'saving restore point for livevolume', $lv->primaryName, $lv, $lv->saveRPoint);
            }

            @rpoints = GREP([qw(sourceVolumeID targetVolumeID activatedVolumeID)], [@ids], (grep { $_->isValid } $self->api->restorePoints(@ssns)));
        }


        if (@rpoints > 1) {

            @rpoints = map { my sc::restorepoint $rp = $_; $rp->isNeedValidation ? $rp->validate : $rp } @rpoints;

            $self->log('find rpoint', 'too many restore points are found', scalar(@rpoints), @rpoints);

            my  @selection = grep { not $_->isErrorState } @rpoints;
                @rpoints = @selection if @selection;
                @selection = grep { $_->isDefinedState } @rpoints;
                @rpoints = @selection if @selection;
        }

        foreach my sc::restorepoint $rp (@rpoints) {

            $rp = $rp->validate if $rp->isNeedValidation;
            $self->log('find rpoint', 'the rpoint is found and selected using volume IDs', $rp);
            $self->update($rp);
            return $rp;
        }
    }

            $self->log('find rpoint', 'failure to find an rpoint', $self->rpid, @ids, @vset);
    return  $self->api->dummy_object(sc::api::TYPE_RestorePoint());
}







#@method
sub resetRPoint {

    my  $self = shift;
        $self->log('reset rpoint', 'trying to reset my rpoint');

    my  sc::restorepoint $rp = shift || 0;

    my  @vset = $self->volumes;

    foreach my sc::volume $v (grep { $_->isLiveVolume } @vset) {

        my sc::livevolume $lv = $v->livevolume;

        if ($lv->isValid) {

            my sc::response $r = $lv->saveRPoint;
            $self->log('reset rpoint', 'saving livevolume rpoint', $r, $lv);
            $self->{VPAIR_Rpoint()} = 0 if $r->success;
            return if $r->success;
        }
    }

    my  @replicas = UNIQ(map { $_->replications } @vset);

    if (@replicas) {

        foreach my sc::replication $repl (@replicas) {

            $self->log('reset rpoint', 'replication found, trying to save new rpoint', $repl);
            my sc::response $r = $repl->saveRPoint;
            $self->log('reset rpoint', 'save new rpoint result', $r);
            $self->{VPAIR_Rpoint()} = 0 if $r->success;
        }
    }
}







#@method
sub activeMappingsRW {

    my  $self = shift;
        $self->log('activemappings(rw)', 'trying to get active write-enabled mappings');

    my  sc::restorepoint $rp = shift || $self->rpoint;
    my  sc::volume $R1 = $rp->sourceVolume;
    my  sc::volume $R2 = $rp->isDRActivated ? $rp->activatedVolume : $rp->targetVolume;
    my  @maps = grep { !$_->isRO } grep { $_ } map { $_->mappings } grep { $_->isUp } grep { $_->isValid } GREPUNIQ(id => $R1, $R2, @_);

        $self->log('activemappings(rw)', 'found '.scalar(@maps).' write-enabled mapping profile(s)', @maps);

    return (@maps);
}







#@method
sub activeMappingsALL {

    my  $self = shift;
        $self->log('activemappings(all)', 'trying to get all mappings (rw+ro)');

    my  sc::restorepoint $rp = shift || $self->rpoint;
    my  sc::volume $R1 = $rp->sourceVolume;
    my  sc::volume $R2 = $rp->isDRActivated ? $rp->activatedVolume : $rp->targetVolume;
    my  @maps = grep { $_ } map { $_->mappings } grep { $_->isUp } grep { $_->isValid } GREPUNIQ(id => $R1, $R2, @_);

        $self->log('activemappings(all)', 'found '.scalar(@maps).' mapping profile(s)', @maps);

    return (@maps);
}







#@method
#@returns sc::volume
sub activeVolume {

    my $self = shift;
       $self->log('activevolume', 'detecting active volume');

    my sc::restorepoint $rpoint = shift || $self->rpoint;


    if ($rpoint->isErrorState) {

        $self->resetRPoint($rpoint);
        $rpoint = $self->findRPoint;
    }


    if ($rpoint->isDRActivated && $rpoint->isTargetAvailable) {

        my sc::volume $av = $rpoint->activatedVolume;
                $self->log('activevolume', 'scenario #2 - DrActivated', $av) if $av->isUp;
        return  $av if $av->isUp;
    }


    if ($rpoint->isSourceFailed) {

        my sc::volume $av = $rpoint->targetVolume;
                $self->log('activevolume', 'scenario #3 - SourceFailed', $av) if $av->isUp;
        return  $av if $av->isUp;
    }


    if ($rpoint->isLiveVolume) {

        if ($self->isLocal($rpoint)) {

            my sc::volume $av = $rpoint->sourceVolume;
            $self->log('activevolume', 'scenario #LV.1 - Primary LV', $av) if $av->isUp;
            return $av if $av->isUp;
        }

        if ($self->isRemote($rpoint)) {

            my sc::volume $av = $rpoint->targetVolume;
            $self->log('activevolume', 'scenario #LV.2 - Secondary LV', $av) if $av->isUp;
            return $av if $av->isUp;
        }
    }

    elsif ($rpoint->isRunningGood) {

        my sc::volume $av = $rpoint->sourceVolume;

        if ($av->isUp) {

            $self->log('activevolume', 'scenario #1 - RunningGood', $av);
            return $av;
        }
    }


    my %disks = (SCAgent::VXDISKS_MAP());
    my @serials = grep { exists($disks{$_}) } $self->serials;


    if (@serials) {

        foreach my sc::volume $av ($self->api->volumes(@serials)) {

            if ($av->isUp) {

                $self->log('activevolume', 'backup method #B1 - an already accessible volume', $av);
                return $av;
            }
        }
    }



    my  @vset = $self->volumes;

    foreach my sc::volume $v (grep { $_->isLiveVolume } @vset) {

        my sc::livevolume $lv = $v->livevolume;

        if ($lv->isUp) {

            if (ISIN($lv->primarySSN, $self->agent->localNodes)) {

                my sc::volume $av = $lv->primary;

                if ($av->isUp) {

                    $self->log('activevolume', 'backup method #LV.B1 - live volume is UP and PRIMARY is on my side', $av);
                    return $av;
                }
            }

            if (ISIN($lv->secondarySSN, $self->agent->localNodes)) {

                my sc::volume $av = $lv->secondary;

                if ($av->isUp) {

                    $self->log('activevolume', 'backup method #LV.B2 - live volume is UP and SECONDARY is on my side', $av);
                    return $av;
                }
            }
        }

        elsif ($lv->isDrActivated) {

            my sc::volume $av = $lv->secondary;

            if ($av->isUp) {

                $self->log('activevolume', 'backup method #LV.B3 - live volume is DRACTIVATED', $av);
                return $av;
            }
        }
    }



    foreach my sc::replication $rp (map { $_->replications } @vset) {

        if ($rp->isUp) {

            my sc::volume $av = $rp->sourceVolume;

            if ($av->isUp) {

                $self->log('activevolume', 'backup method #B2 - replication is UP, source volume', $av);
                return $av;
            }
        }
    }



    if ($self->swap) {

        my sc::volume $av = $self->api->volumes($self->swap);

        if ($av->isUp) {

            $self->log('activevolume', 'backup method #B3 - by registered swap serial', $av);
            return $av;
        }
    }



    #anything else - is a moot point, and in such case the Volume should not be automatically mapped to a server
            $self->log('active volume', 'unclear scenario, return an empty object', $rpoint, {%$rpoint}, {%$self});
            $self->resetRPoint($rpoint);
    return  $self->api->dummy_object(sc::api::TYPE_Volume);
}







#@method
sub drConfig {

    my $self = shift;
    my sc::restorepoint $rp = shift || $self->rpoint;
    return LOG(MSG_VPAIR_RPDEF_ERR_INVALID, $rp) unless $rp->isValid;

    my @servers = $self->agent->servers($rp->targetSSN);
    $self->trace('dr activation config', 'agent servers listing', @servers);
    return LOG(MSG_VPAIR_RPDEF_ERR_SERVER,  $rp, $self->agent->servers) unless @servers;

    my $result = {

        sc::api::PARAM_RPDR_Name()                          => $rp->targetVolumeName,
        sc::api::PARAM_RPDR_PreserveLiveVolume()            => 1,
        sc::api::PARAM_RPDR_SyncReplicationBeforeActivate() => 1,
        sc::api::PARAM_RPDR_UseActiveReplay()               => 1,
        sc::api::PARAM_RPDR_UseOriginalVolumesFolder()      => 1,
        sc::api::PARAM_RPDR_ServerList()                    => [ map {$_->id} @servers ]
    };

    $self->log('dr activation config', 'new settings', $result);
    return $result;
}







#@method
sub configRestorePoint {

    my  $self = shift;
    my  sc::restorepoint $rp = shift || $self->rpoint;
        $self->log('predefine', 'restore point', $rp);

    if (my $config = $self->drConfig($rp)) {

        my  $resp = $rp->predefine($config);
            $self->log('predefine', 'api answer', $resp);

        if ($resp->isa('sc::response') && $resp->success) {

            $rp->validate;
            $self->log('predefine', 'new settings', $rp->settings);
            return 1;
        }

        return LOG(MSG_VPAIR_RPDEF_ERR_FAILED,  $resp)
    }

    return 0;
}







#@method
sub set_swap {

    my              $self   = shift;
    my  sc::volume  $vol    = shift;
    my              $old    = $self->swap;

    $self->log('initSwap', 'swap requested', $vol);
    return 0 unless ($vol && ref($vol) && $vol->isa('sc::volume') && $vol->isUp && $vol->serial);

    LOG(MSG_VPAIR_SWAP_STARTED(), $vol, $self);
    $self->{VPAIR_Swap()} = $vol->serial;
    return 1 if $self->agent->updateVpair($self);

    LOG(MSG_VPAIR_SWAP_CANCELED(), $vol, $self);
    $self->{VPAIR_Swap()} = $old;
    return 0;
}







#@method
sub wait_swap {

    my $self = shift;
    return 0 unless $self->swap;

    $self->log('wait_swap', 'trying to commit swap', $self->swap);


    my sc::volume $drv = $self->api->volumes($self->swap);
    my sc::restorepoint $rp = 0;


    if ($drv->isUp && $drv->isSource) {

        foreach my sc::replication $r ($drv->replications) {

            if ($r->isUp) {

                $rp = $r->restorepoint;

                unless ($rp->isValid) {

                    $r->saveRPoint;
                    $rp = $r->restorepoint;
                }

                last if $rp->isRunningGood;
            }
            else {

                $self->log('wait_swap', 'replication exists but is not ready for swap', $r);
            }
        }
    }



    unless ($rp && $rp->isDefinedState) {

        return 1 if $self->retry(3);
                    $self->resetRPoint($rp);
        $rp =       $self->findRPoint;
        return 0 unless $self->swap;
    }



    unless ($rp && $rp->isRunningGood) {

        $self->log('wait_swap', 'rpoint is not ready', $rp);
        return 1 if $self->retry(3);
    }



    my sc::volume $sv = $rp->sourceVolume;
    my sc::volume $tv = $rp->targetVolume;

    unless ($sv->isUp && $tv->isUp) {

        $self->log('wait_swap', 'volumes are not ready', $sv, $tv);
        return 1 if $self->retry(3);
    }


    LOG MSG_VPAIR_SWAP_READY, $self, $rp, $sv, $tv;
    return 0 if (not $self->swap) || $self->update($rp);
    return 1 if $self->logAndRetry('wait_swap', 'failure to update vpair, will retry', $rp);

    LOG MSG_VPAIR_SWAP_FAILED, $self, $rp, $sv, $tv;
    return 0;
}







#@method
#@returns sc::drtask
sub task {

    my              $self = shift;
    my sc::drtask   $t = $self->{VPAIR_Task()};
    return          $self->api->dummy_object(sc::api::TYPE_DrTask()) unless ($t && ref($t) && $t->isa('sc::drtask'));
                    $t = $t->update;
    return          $self->{VPAIR_Task()} = $t if ($t && ref($t) && $t->isa('sc::drtask') && $t->isValid);
    return          $self->api->dummy_object(sc::api::TYPE_DrTask());
}







#@method
#@returns sc::drtask
sub set_task {

    my $self = shift;
    my sc::task $t = shift;

    if ($t && ref($t) && $t->isa('sc::task')) {

        return  $self->{VPAIR_Task()} = $t if $t->isRunning;
                $self->trace('set_task', 'the task is not Running and therefor cannot be registered', $self, $t)
    }
    else {
            $self->trace('set_task', 'attempt to assign invalid task', $self, $t);
    }

    return  $self->api->dummy_object(sc::api::TYPE_DrTask());
}







#@method
sub wait_task {

    my          $self   = shift;
    my sc::task $task   = $self->task;

    if ($task && $task->isValid) {

        $self->log('wait_task', $task->toString(qw(method state stepMessage)), $task, {%$task});

        if ($task->isRunning) {

            LOG(MSG_VPAIR_TASK_RUNNING, $self, $task);

            if ($task->isa('sc::drtask')) {

                my sc::drtask $drt = $task;
                LOG(MSG_VPAIR_NEEDDEACTIVATE, $self, $drt->deactivateDR) if ($drt->needDeactivate);
            }
        }
        else {

            LOG(MSG_VPAIR_TASK_FAILED,      $self, $task) if $task->isError;
            LOG(MSG_VPAIR_TASK_COMPLETE,    $self, $task) unless $task->isError;
            $self->{VPAIR_Task()} = 0;
        }

        return 1;
    }

    return 0;
}







#@method
sub online {

    my  $self = shift;
        $self->trace('online', 'starting decision cycle', $self);

    return 0 if $self->wait_task;



    my sc::restorepoint $rp = $self->rpoint;

    unless ($rp->isValid) {

        if ($self->retry) {

            $self->log('online', 'rpoint is invalid', $self->{VPAIR_Rpoint()}, {%$rp});
            $self->resetRPoint;
            return 0;
        }

        $self->log('online', 'rpoint is still invalid, give up');
        $self->trace('online', 'failure to find rpoint. Here is the list of all available RPoints in the array', map {{ %$_ }} $self->api->restorePoints);
        LOG MSG_ONLINE_ERROR, $self, $rp;
        return 1;
    }


    my $state  = $rp->classify;





    if ($state == sc::restorepoint::STATE_01_RunningGood()) {

        LOG MSG_VPAIR_RUNNINGGOOD, $self, $rp; $self->log('online', 'state = running good');

        if ($rp->isLiveVolume) {

            my sc::livevolume $lv = $rp->liveVolume;

            if ($lv->isUp) {

                if (ISIN($lv->primarySSN, $self->agent->localNodes)) {

                    $self->log('online', 'decision = good to go');
                    LOG MSG_ONLINE_GOODTOGO, $self, $lv;
                    return 1;
                }

                elsif (ISIN($lv->secondarySSN, $self->agent->localNodes)) {

                    $self->log('online', 'decision = swap livevolume');
                    my sc::task     $task = $lv->swap;
                    return          $self->logAndRetry('online', 'failure to start livevolume swap', $task, {%$task}) unless $task->isRunning;
                    LOG MSG_ONLINE_WAIT_LV_SWAP, $self, $lv, $self->set_task($task);
                    return 0;
                }

                else {

                    $self->log('online', 'decision = cannot do much here, go away from this node');
                    LOG MSG_ONLINE_GO_AWAY_INVISIBLE, $self, {%$rp}, {'local nodes' => [$self->agent->localNodes]};
                    return 1;
                }
            }

            else {

                return $self->logAndRetry('online', 'livevolume is unhealthy or not ready yet', $lv);
            }
        }

        else {

            if ($self->isLocal($rp)) {

                $self->log('online', 'decision = good to go');
                LOG MSG_ONLINE_GOODTOGO, $self, $rp;
                return 0 if $self->wait_swap;
                return 1;
            }

            else {

                                $self->log('online', 'decision = activate dr volume');
                                $self->configRestorePoint($rp);
                                $rp->validate;
                my sc::drtask   $task = $rp->activateDR();
                return          $self->logAndRetry('online', 'failure to start dr activation', $task, {%$task}) unless $task->isRunning;
                LOG MSG_ONLINE_WAIT_DRACTIVATE, $self, $self->set_task($task);
                return 0;
            }
        }
    }





    elsif ($state == sc::restorepoint::STATE_02_SourceFailed()) {

        LOG MSG_VPAIR_SOURCEFAILED, $self, $rp; $self->log('online', 'state = source is failed');

        if ($self->isRemote($rp)) {

                            $self->log('online', 'decision = activate dr volume');
                            $self->configRestorePoint($rp);
                            $rp->validate;
            my sc::drtask   $task = $rp->activateDR();
            return          $self->logAndRetry('online', 'failure to dr activation', $task, {%$task}) unless $task->isRunning;
            LOG MSG_ONLINE_WAIT_DRACTIVATE, $self, $self->set_task($task);
            return 0;
        }

        else {

            $self->log('online', 'decision = cannot do much here, go away from this node');
            LOG MSG_ONLINE_GO_AWAY_SOURCE, $self, {%$rp}, {'local nodes' => [$self->agent->localNodes]};
            return 1;
        }
    }





    elsif ($state == sc::restorepoint::STATE_03_FailoverInProgress()) {

        LOG MSG_VPAIR_DRACTIVATING, $self, $rp; $self->log('online', 'state = dr is activating...');

        my @tasks = $rp->drTasks(sc::api::STATE_RUNNING, sc::api::STATE_ERROR);

        foreach my sc::drtask $t (@tasks) {

            $self->log('online', 'dr task found', $t, {%$t});
        }

        unless (scalar(grep {$_->isRunning} @tasks)) {

                    $self->resetRPoint($rp);
            return  $self->logAndRetry('online', 'looks like the dr activate job is hung, trying to reset rpoint', $#tasks, @tasks);
        }

        return 0;
    }





    elsif ($state == sc::restorepoint::STATE_04_TargetActivated()) {

        LOG MSG_VPAIR_DRACTIVATED, $self, $rp; $self->log('online', 'state = DrActivated');

        if ($rp->isLiveVolume) {

            my sc::livevolume $lv = $rp->liveVolume;

            if ($lv->isValid && $lv->isRepairable) {

                my sc::drtask $task;

                        $self->log('online', 'decision = restore LiveVolume', $lv);
                        $task = $rp->activateRestore;
                return  $self->logAndRetry('online', 'failure to start LiveVolume restore', $task, $rp) unless $task->isRunning;

                LOG MSG_ONLINE_WAIT_RESTORE, $self, $self->set_task($task);
                return 0;
            }
        }

        my sc::volume $sv = $rp->sourceVolume;

        unless ($sv->isUp) {

            $self->log('online', 'failure to get source volume', $sv, { %$sv }) unless $sv->isValid;
            $self->log('online', 'source volume is down', $sv, { %$sv }) if $sv->isValid;

            if ($self->isRemote($rp)) {

                $self->log('online', 'decision = good-to-go');
                LOG MSG_ONLINE_GOODTOGO, $self;
                return 1;
            }
            else {

                $self->log('online', 'decision = cannot do much here, go away from this node');
                LOG MSG_ONLINE_GO_AWAY_TARGET, $self, { %$rp }, { 'local nodes' => [ $self->agent->localNodes ] };
                return 1;
            }
        }

        elsif ($self->isLocal($rp) and not $self->isRemote($rp)) {

            $self->log('online', 'decision = restore back to source, cause this host wont be able to access the target volume');
            my sc::drtask   $task = $rp->activateRestore;
            return $self->logAndRetry('online', 'failure to start restore', $task, { %$task }) unless $task->isRunning;
            LOG MSG_ONLINE_WAIT_RESTORE, $self, $self->set_task($task);
            return 0;
        }

        elsif ($rp->isTargetAvailable) {

            my sc::volume $drv = $rp->activatedVolume;

            if ($drv->isUp) {

                $self->log('online', 'decision = good-to-go, target volume is available');
                LOG MSG_ONLINE_GOODTOGO, $self;
                return 1;
            }
            else {

                return $self->logAndRetry('online', 'dr activated volume is not available', $drv);
            }
        }

        elsif ($self->isLocal($rp)) {

            $self->log('online', 'decision = restore back to source, cause target volume seems to be unavailable as per restore point information');
            my sc::drtask   $task = $rp->activateRestore;
            return $self->logAndRetry('online', 'failure to start restore', $task, { %$task }) unless $task->isRunning;
            LOG MSG_ONLINE_WAIT_RESTORE, $self, $self->set_task($task);
            return 0;
        }

        return $self->logAndRetry('online', 'dr is activated, but cannot detect a valid scenario', $rp);
    }





    elsif ($state == sc::restorepoint::STATE_05_RestoreInProgress()) {

        LOG MSG_VPAIR_RESTORING, $self, $rp; $self->log('online', 'state = restoring');

        my @tasks = $rp->drTasks(sc::api::STATE_RUNNING, sc::api::STATE_ERROR);

        foreach my sc::drtask $t ($rp->drTasks(sc::api::STATE_RUNNING, sc::api::STATE_ERROR)) {

            $self->log('online', 'dr task found', $t, {%$t});
            $self->log('online', 'need to deactivate target', $t) if $t->needDeactivate;
            LOG (MSG_VPAIR_NEEDDEACTIVATE, $self, $self->set_task($t), $t->deactivateDR) if $t->needDeactivate;
        }

        unless (scalar(grep {$_->isRunning} @tasks)) {

                    $self->resetRPoint($rp);
            return  $self->logAndRetry('online', 'looks like the restore job is hung, trying to reset rpoint', $#tasks, @tasks);
        }

        return 0;
    }





    elsif ($state == sc::restorepoint::STATE_00_Error()) {

        if ($self->retry(3)) {

            return $self->log('online', 'rpoint is in ERROR state', $rp, {%$rp});
            return 0;
        }
        elsif ($self->retry) {

            $self->resetRPoint($rp);
            return $self->log('online', 'rpoint is in ERROR state', $rp, {%$rp});
            return 0;
        }

        LOG MSG_ONLINE_ERROR, $self, $rp;
        return 1;
    }





    if ($self->retry(3)) {

        return $self->log('online', 'rpoint is in UNKNOWN state', $rp, {%$rp});
        return 0;
    }

    elsif ($self->retry) {

        $self->resetRPoint($rp);
        return $self->log('online', 'rpoint is in UNKNOWN state', $rp, {%$rp});
        return 0;
    }

    LOG MSG_ONLINE_UNKNOWN, $self, $rp;
    return 1;
}







#@method
sub reverse {

    my  $self = shift;
        $self->trace('reverse', 'starting decision cycle', $self);
        return 0 if $self->wait_task;
        return 0 if $self->wait_swap;


    my sc::restorepoint $rp = $self->rpoint;
        return $self->logAndRetry('reverse', 'rpoint is not valid', $rp) unless $rp->isValid;

    if ($rp->isLiveVolume) {

        $self->log('reverse', 'decision = good-to-go, cause livevolumes do not need reverse operation');
        return 1
    }

    unless ($rp->isDefinedState) {

        return $self->logAndRetry('reverse', 'waiting for rpoint to get into defined state', $rp);
    }





    my  $state  = $rp->classify;



    if ($state == sc::restorepoint::STATE_01_RunningGood()) {

        $self->log('reverse', 'state = running good');
        $self->log('reverse', 'decision = good-to-go');
        return 1;
    }



    elsif ($state == sc::restorepoint::STATE_04_TargetActivated()) {

        $self->log('reverse', 'state = DrActivated');

        if ($rp->isTargetAvailable) {

            my sc::volume $sv  = $rp->sourceVolume;
            my sc::volume $drv = $rp->activatedVolume;
            my sc::drtask $task;

            return  $self->logAndRetry('reverse', 'source volume seems to be down', $sv, {%$sv})            unless $sv->isUp;
            return  $self->logAndRetry('reverse', 'dr-activated volume seems to be down', $drv, {%$drv})    unless $drv->isUp;
            return  $self->logAndRetry('reverse', 'failure to register swap attempt')                       unless $self->set_swap($drv);
                    $self->log('reverse', 'decision = mirror back');
                    $task  = $self->set_task($rp->activateMirrorBack);

            if ($task->isRunning) {

                LOG MSG_ONLINE_WAIT_MIRRBACK, $self, $task;
                return 0;
            }

            return  $self->logAndRetry('reverse', 'swap is canceled, cause mirror back request was unsuccessful', $self->api->lastResponse)
        }

        $self->log('reverse', 'dr is activated, but destination volume is not available; leaving the volume pair as it is');
        return 1;
    }



    elsif  ($state == sc::restorepoint::STATE_05_RestoreInProgress()) {

            $self->log('reverse', 'state = restoring');

        my  @tasks = $rp->drTasks(sc::api::STATE_RUNNING, sc::api::STATE_ERROR);

        foreach my sc::drtask $t (@tasks) {

            $self->log('reverse', 'task found', $t, {%$t});
            $self->log('reverse', 'need to deactivate target') if $t->needDeactivate;
            LOG(MSG_VPAIR_NEEDDEACTIVATE, $self, $self->set_task($t), $t->deactivateDR) if $t->needDeactivate;
        }

        return 0 if scalar(grep {$_->isRunning} @tasks);

        if ($self->retry(3)) {

            $self->log('reverse', 'restore job seems to be stuck, trying to restart restore');

            if ($rp->isTargetAvailable) {

                my sc::volume $sv  = $rp->sourceVolume;
                my sc::volume $drv = $rp->activatedVolume;
                my sc::drtask $task;

                if ($sv->isUp && $drv->isUp) {

                    $self->log('reverse', 'decision = retry mirror back');
                    $task = $rp->activateMirrorBack;
                    $self->log('reverse', 'mirror back task', $task);
                    $self->set_task($task) if $task->isRunning;
                    return 0;
                }
            }

            $self->log('reverse', 'failure to restart restore job cause source volume and(or) target volume are not available; leaving the volume pair as it is');
            return 1;
        }

        $self->log('reverse', 'restore job is hung, restore reset doesnt help; leaving volume pair as it is');
        return 1;
    }



    elsif ($state == sc::restorepoint::STATE_02_SourceFailed()) {

        $self->log('reverse', 'state = source is failed');
        $self->log('reverse', 'decision = cannot do much here, cause SourceFailed condition should have been treated at the Online stage');
        return 1;
    }



    elsif ($state == sc::restorepoint::STATE_03_FailoverInProgress()) {

        $self->log('reverse', 'state = dr is activating...');
        $self->log('reverse', 'decision = cannot do much here, cause FailoverInProgress condition should have been treated at the Online stage');
        return 1;
    }



    elsif ($state == sc::restorepoint::STATE_00_Error()) {

        return $self->logAndRetry('reverse', 'rpoint is en error state', $rp);
    }


    return $self->logAndRetry('reverse', 'rpoint is an unknown state', $rp);
}







#@method
sub mapping {

    my  $self = shift; $self->log('mapping', 'ensure host access');
    my  $disk = shift || { SCAgent::VXDISKS_MAP() };
    my  sc::volume $volume = $self->activeVolume;

    if (scalar(grep {exists($disk->{$_})} $self->serials)) {

        $self->log('mapping', 'my serial number is already visible on the host');
    }

    unless ($volume->isUp) {

        return $self->logAndRetry('mapping', 'volume is not ready for mapping', $volume);
    }

    if (exists($disk->{$volume->serial})) {

        $self->log('mapping', 'active volume serial (' . $volume->serial . ') is already visible on the host', $disk->{$volume->serial}, $volume);

        foreach my sc::mappingprofile $m ($self->activeMappingsALL(0, $volume)) {

            $self->log('mapping', 'ensuring mapping consistency', $m);
            $m->makeRW if $m->isRO && $m->volumeID eq $volume->id;
            $m->makeRO unless $m->volumeID eq $volume->id;
        }

        return 1;
    }

    unless ($self->agent->isAttachedTo($volume->ssn)) {

        $self->log('mapping', 'active volume can not be visible on this host because the host is not connected to the appropriate SC-node (SSN: ' . $volume->ssn. ')', $volume, $self->agent->localNodes);
        return 1;
    }


    my  @mappings = $volume->mappings;
        $self->trace('mapping', 'current mappings', $volume, @mappings);

    my  %servers  = map { $_->id => $_ } $self->agent->servers($volume->ssn);
        $self->trace('mapping', 'expected servers', {%servers});

    foreach my sc::mappingprofile $m (@mappings) {

        if ($m->isUp) {

            if (exists($servers{$m->serverID})) {

                $m->makeRW if $m->isRO;
                delete($servers{$m->serverID});
            }

            else {

                $m->makeRO unless $m->isRO;
            }
        }

        else {

            $self->log('mapping', 'removing invalid map', $m, $m->delete);
        }
    }

    unless (scalar(%servers)) {

        $self->log('mapping', 'mapping is all good');
        return 0;
    }

    foreach my $id (keys %servers) {

        $self->log('mapping', 're-mapping the volume', $volume, $servers{$id}, $volume->mapToServer($id));
    }

    return 0 if $self->retry;

        $self->log('mapping', 'no valid mapping found, give up');

    return 1;
}





1;